"""States used in the ``django_instances_graph`` module.

Attributes
----------
    STARTED: str
        Value used when a pass on deserializing an object is done but the object is not saved, or
          relations are not all done yet
    RUNNING: str
        Value used when an operation is currently being processed
    SAVED: str
        Value used when a deserializing object is saved in DB but fields other not direct fields
        (ie simple and FKs) are not all done yet
    DONE: str
        Value used when an operation is finished


"""

STARTED = 'started'
RUNNING = 'running'
SAVED = 'saved'
DONE = 'done'
