"""
``django_instances_graph`` provides a way to get the whole graph of instances starting from one.
"""

from .graph import Graph
from .instance import Instance


__author__ = 'Stephane "Twidi" Angel'
__contact__ = 's.angel@twidi.com'
__homepage__ = 'https://github.com/magencydigital/'
__license__ = 'BSD'
__version__ = '1.1.1'
VERSION = tuple(__version__.split('.'))

